/*

** MyGif.c

**

**   Stupid & simple EGA GIF viewer.  Does not load color palette, simply

**   decodes and displays the image.

**

**   Rewrite LoadPalette() to perform palette loading.  I didn't insert

**   this since my viewer uses my frame grabber board for display, and

**   the palette loading code is very different.  Also loading a VGA

**   palette is different from that of the EGA.  Customize for your

**   system.

**

**   I hereby release this code to the public domain, no restrictions,

**   no warranty, no nothing.  This applies only to this file (mygif.c),

**   see the other files for other copyrights, etc.

*/



#include <sys/types.h>

#include <sys/stat.h>

#include <conio.h>

#include <fcntl.h>

#include <io.h>

#include <stdio.h>

#include <malloc.h>

#include <string.h>

#include <graph.h>



#include "errs.h"



#define FALSE    0

#define TRUE     1



#define min(x,y) (x<y ? x:y)

#define max(x,y) (x>y ? x:y)



#define ROWCNT   350

#define COLCNT   640



typedef unsigned int  uint;

typedef unsigned char uchar;





/*

 VDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD7

 : Working variables - globals, screen info, color table            :

 SDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDDD=

*/



int         infile;             /* Input file handle               */

uint        screenwidth;        /* The dimensions of the screen    */

uint        screenheight;       /*   (not those of the image)      */

short       global;             /* Is there a global color map?    */

int         globalbits;         /* Number of bits of global colors */

int         backclr;            /* Background color                */

int         colres;             /* Number of bits of color resol.  */

int        *interleavetable;

uchar       globalmap[256][3];  /* RGB values for global color map */

uchar       colortable[256*3];  /* RGB values for image            */



uint        row,left,top, width,height;



void        usage(), fatal(), checksignature(), readextension();

void        readscreen(), initcolors(), readimage();







/* Support routines: customize for your system

**         get_byte() - return next byte of GIF image data

**         out_line() - display a line of graphics on the screen

**

**         Load_Pallete() - load the proper colors into the palette

*/





get_byte() {

   uint ch;



   /* VERY slow, use buffered reading for faster throughput */



   if (read(infile,(char *)&ch,1)==1) return(ch);

   else                               return(READ_ERROR);

}





out_line(line, len)

uchar line[];

int   len;

{

   register uint i;



   /* slow, maybe write out entire line at once, instead of pixel by pixel */



   if (interleavetable[row] < ROWCNT) {

      for(i=0; i<min(len,COLCNT); i++) {

         _setcolor(line[i]);

         _setpixel(i,interleavetable[row]);

      }

   }



   row++;

}





LoadPalette() {

   /* Colors are in global 'colortable', first are 256 red colors, 256 green

   ** and then 256 blue, I think :-).   Not needed for program to run, but

   ** the colors will not be correct.

   **

   ** This code will be different for Ega, Vga, Suns, etc, so roll your own.

   ** (Mine was written for my frame grabber, it wouldn't help you).

   */

}

















/*

** Main Program - open specified GIF file, read and verify GIF signature

**                load screen descriptor, global color map

**                        Z load image descriptor ?

**                Repeat -3 load local color map  3- for each image in file

**                        3 decode image          3

**                        @ skip extension blocks Y

**                Done - pause for keypress

*/



main(argc,argv)

int argc;

char *argv[];

{

        int  quit;

        char ch;



        if (argc!=2) usage();



        _setvideomode(_DEFAULTMODE);

        if ((infile = open(argv[1],(O_BINARY|O_RDONLY)))==-1) {

              perror("mygif");

              exit(-1);

           }



        checksignature();

        readscreen();



        quit = FALSE;

        do {

            ch = get_byte();

            switch (ch) {

                case '\0':  break;  /* this kludge for non-standard files */

                case ',':   readimage();

                            break;

                case ';':   quit = TRUE;

                            break;

                case '!':   readextension();

                            break;

                default:    break;

            }

        } while (!quit);



        while( !kbhit() );



        _setvideomode(_DEFAULTMODE);

}





void usage() {

        fprintf(stderr,"usage: mygif [GIF-file]\n");

        exit(-1);

}





void fatal(s) char *s; {

        fprintf(stderr,"mygif: %s\n",s);

        exit(-1);

}





void checksignature() {

        char buf[6];



        read(infile,buf,6);

        if (strncmp(buf,"GIF",3)) fatal("file is not a GIF file");

        if (strncmp(&buf[3],"87a",3)) fatal("unknown GIF version number");

}





void readscreen() {

        uchar buf[7];



        read(infile,buf,7);



        if (buf[6]) fatal("screen descriptor is invalid");



        screenwidth  = buf[0] + (buf[1] << 8);

        screenheight = buf[2] + (buf[3] << 8);

        global       = buf[4] & 0x80;

        colres       = ((buf[4] & 0x70) >> 4) +1;

        backclr      = buf[5];



        if (global) {

            globalbits = (buf[4] & 0x07) + 1;

            read(infile,(char *)globalmap,3*(1<<globalbits));

        }

}





void initcolors(colortable,colormap,ncolors)

uchar colortable[256*3];

uchar colormap[256][3];

int ncolors;

{

        register int entry, rgb;

        register uchar *cp;



        cp = colortable;

        for (rgb = 0; rgb < 3 ; rgb++)

                for (entry = 0; entry < ncolors ; entry++)

                        *(cp)++ = colormap[entry][rgb];

}





void readimage()

{

        uchar    buf[9];

        short    local,interleaved;

        char     localmap[256][3];

        int      localbits, ncolors;

        register i;



        read(infile,buf,9);

        left        = buf[0] + (buf[1] << 8);

        top         = buf[2] + (buf[3] << 8);

        width       = buf[4] + (buf[5] << 8);

        height      = buf[6] + (buf[7] << 8);

        interleaved = buf[8] & 0x40;



        if (local = (buf[8] & 0x80)) {

            localbits = (buf[8] & 0x7) + 1;

            ncolors = 1 << localbits;

            read(infile,(char *)localmap,3*ncolors);

            initcolors(colortable,localmap,ncolors);

        } else if (global) {

            initcolors(colortable,globalmap,ncolors = (1 << globalbits));

        } else

            fatal("no colormap present for image");



        printf("GIF87a  %dx%d %s lzw encoded bitmap @ (%d, %d)\n",width,height,(interleaved ? "interleaved":""),left,top);

        if (global) printf("Global Map Present, %d color bits, %d color res\n", globalbits, colres);

        if (local)  printf("Local  Map Present, %d color bits\n", localbits);



        _setvideomode(_ERESCOLOR);

        _clearscreen(_GCLEARSCREEN);

        LoadPalette(colortable);



        if (!(interleavetable = (int*)malloc(height*sizeof(int))))

           fatal("not enough memory for interleave table");



        if (interleaved) {

            row = 0;

            for (i = 0; i < height; i += 8) interleavetable[row++] = top+i;

            for (i = 4; i < height; i += 8) interleavetable[row++] = top+i;

            for (i = 2; i < height; i += 4) interleavetable[row++] = top+i;

            for (i = 1; i < height; i += 2) interleavetable[row++] = top+i;

        } else

            for (i = 0; i < height; i++) interleavetable[i] = top+i;



        row = 0;

        switch(decoder(width)) {

           case(OUT_OF_MEMORY): fatal("Out of memory");

                                break;

           case(BAD_CODE_SIZE): fatal("Bad Code Size");

                                break;

           case(READ_ERROR):    fatal("Read Error");

                                break;

           case(WRITE_ERROR):   fatal("Write Error");

                                break;

           case(OPEN_ERROR):    fatal("Open Error");

                                break;

           case(CREATE_ERROR):  fatal("Create Error");

                                break;

        }



        if (interleaved) free(interleavetable);

}





void readextension() {

        uchar code,count;

        char buf[255];



        code = get_byte();

        while (count = get_byte()) read(infile,buf,count);

}





/*

** End of MyGif.c

*/



